/* $Id: lifedraw.c,v 1.10 2005/01/25 19:45:21 muli Exp $ */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <curses.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "linux/klife.h"

#include "version.h"

static const int rpentomino[][2] = {
	{13, 14} , {13, 15}, {14, 13}, {14, 14}, {15, 14}, {-1, -1}
};
static const int glider[][2] = {
	{13, 14}, {14, 14}, {14, 15}, {15, 13}, {15, 15}, {-1, -1}
};

static int gkfd;

struct klife_opts {
	int mmap;
	int timer;
	void* initial_config;
	size_t initial_config_size;
	void* ptr; /* mmap */
};

static void finish(int kfd, int status);

static void usage(const char* prog, int status)
{
        const char* str =
                "klife version %s.\n\n"
		"Usage: \n"
                "%s [-g|--glider] [-m|--mmap] [-t|--timer], or\n"
		"%s [-r|--rpentomino] [-m|--mmap] [-t|--timer]\n";

        fprintf(stderr, str, version, prog, prog);
        exit(status);
}

static int write_initial_positions(int kfd, const int starting[][2], size_t sz)
{
	const int* pair = starting[0];
	char buf[32];
	int ret;

	while (pair[0] != -1 && pair[1] != -1) {
		size_t sz = sizeof(pair[0]);
		memcpy(buf, &pair[0], sz);
		memcpy(buf + sz, &pair[1], sz);
		sz *= 2;
		if ((ret = write(kfd, buf, sz) != sz)) {
			fprintf(stderr, "%s: write returned %d\n",
				__func__, ret);
			if (ret < 0)
				return ret;
			return -EPROTO;
		}
		pair += 2;
	}
	return 0;
}

static int
kfd_setup(struct klife_opts* opts)
{
	int kfd;
	int ret;
	int timer;
	void* ptr;

	if ((kfd = open("/dev/klife", O_RDWR)) < 0) {
		perror("open /dev/klife");
		exit(EXIT_FAILURE);
	}

	ret = write_initial_positions(kfd, opts->initial_config, 
				      opts->initial_config_size);
	if (ret < 0)
		goto out;

	if (opts->mmap) {
		ptr = mmap(NULL, TOTAL_GRID_SIZE, PROT_READ, MAP_PRIVATE, kfd, 0);
		if (ptr == MAP_FAILED) {
			fprintf(stderr, "mmap failed, errno %s\n", strerror(errno));
			goto out;
		}
		opts->ptr = ptr;
	}

	timer = opts->timer;
	if ((ret = ioctl(kfd, KLIFE_SET_TIMER_MODE, &timer)) < 0) {
		fprintf(stderr, "ioctl(KLIFE_SET_TIMER_MODE) returned %d "
			"(errno %s)\n", ret, strerror(errno));
		goto out;
	}

	gkfd = kfd; /* sighandler needs access to it ... */
	return kfd;
out:
	/* undo an mmap, if necesary */
	if (opts->ptr)
		munmap(ptr, TOTAL_GRID_SIZE);

	close(kfd);
	exit(EXIT_FAILURE);
}

static size_t
generation(int kfd, char* buffer, size_t sz, unsigned long* gen,
	   struct klife_opts* opts)
{
	int ret;

	if ((ret = read(kfd, buffer, sz)) < 0) {
		perror("read");
		finish(kfd, EXIT_FAILURE);
	}

	if (ret == 0 && !opts->mmap) {
		fprintf(stderr, "read returned EOF?!\n");
		finish(kfd, EXIT_FAILURE);
	}

	if ((ret = ioctl(kfd, KLIFE_GET_GENERATION, gen)) < 0) {
		fprintf(stderr, "ioctl(KLIFE_GET_GENERATION) returned %d "
			"(errno %s)\n", ret, strerror(errno));
		finish(kfd, EXIT_FAILURE);
	}

	return ret;
}

static void draw_mapped(struct klife_opts* opts)
{
	int i;
	int j;
	char c;
	struct grid* grid = opts->ptr;

	for (i = 0; i < GRID_SIZE; ++i) {
		for (j = 0; j < GRID_SIZE; ++j) {
			c = cell_is_alive(grid, i, j) ? 'X' : ' ';
			/* keep the first two rows clear */
			mvaddch(i + 1, j, c);
		}
	}
}

static void
draw(const char* buffer, size_t len, unsigned long gen, struct klife_opts* opts)
{
	char genstr[80];

	snprintf(genstr, sizeof(genstr), 
		 "generation: %ld%s%s", 
		 gen, opts->timer ? " [timer mode]" : "",
		 opts->mmap ? " [mmap mode]" : "");

	mvaddstr(0, 0, genstr);

	if (opts->mmap)
		draw_mapped(opts);
	else
		mvaddstr(1, 0, buffer);

	refresh();
}

static void finish_sighand(int sig)
{
	finish(gkfd, EXIT_SUCCESS);
}

static void finish(int kfd, int status)
{
	endwin();
	close(kfd);
	exit(status);
}

static void curses_init(void)
{
	/* initialize your non-curses data structures here */
	(void)signal(SIGINT, finish_sighand);  /* arrange interrupts to terminate */

	(void)initscr(); /* initialize the curses library */
	keypad(stdscr, TRUE);  /* enable keyboard mapping */
	(void)nonl(); /* tell curses not to do NL->CR/NL on output */
	(void)cbreak(); /* take input chars one at a time, no wait for \n */
	(void)noecho(); /* don't echo input */
}

static void
parse_command_line_args(int argc, char** argv, struct klife_opts* opts)
{
        int option_index = 0;
        char* prog = argv[0];

	memset(opts, 0, sizeof(*opts));

	if (argc < 2)
		usage(prog, EINVAL);

        while(1){
                /* structure with all recognised options for sct_config */
                static struct option long_options[] = {
                        {"timer", 0, NULL, 't'},
                        {"mmap", 0, NULL, 'm'},
                        {"rpentomino", 0, NULL, 'r'},
                        {"glider", 0, NULL, 'g'},
                        {0, 0, NULL, 0}
                };

                int c;
                c = getopt_long (argc, argv, "+tmrg", long_options, &option_index);

                switch (c) {
                case 't':
			opts->timer = 1;
                        break;
                case 'm':
			opts->mmap = 1;
                        break;
                case 'r':
			if (opts->initial_config)
				usage(prog, EINVAL);

			opts->initial_config = rpentomino;
			opts->initial_config_size = sizeof(rpentomino);
                        break;
                case 'g':
			if (opts->initial_config)
				usage(prog, EINVAL);

			opts->initial_config = glider;
			opts->initial_config_size = sizeof(glider);
                        break;
                case '?': /* unrecognised option */
                        /* fall through */
                default:
                        c = -1;
                        break;
                }
                if(c == -1)
                        break;  /* no more options */
        }

        argc -= optind;
        if (argc)  /* there should be no commands remaining */
                usage(prog, EINVAL);

	if (!opts->initial_config)
		usage(prog, EINVAL);
}

int main(int argc, char* argv[])
{
	int kfd;
	unsigned long gen;

	char buf[4096];
	size_t sz;

	struct klife_opts opts;

	parse_command_line_args(argc, argv, &opts);

	kfd = kfd_setup(&opts);

	curses_init();

	while (1) {
		sz = generation(kfd, buf, sizeof(buf), &gen, &opts);
		draw(buf, sz, gen, &opts);
		sleep(1);
	}

	finish(kfd, EXIT_SUCCESS);
}
